import polars as pl

from ..interface import SourceInterface
from ..registrated_source import register_source
from .excel_config import ExcelConfig, ExcelConfigArguments


@register_source("Excel")
class ExcelSource(SourceInterface):
    ConfigArgumentsClass = ExcelConfigArguments
    ConfigClass = ExcelConfig

    """ A source for static data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        return pl.read_excel(
            self.connection_settings.path,
        )

    def insert(self, data: pl.DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: pl.DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: pl.DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)



from polars import DataFrame


class ComparisonResult:
    create: DataFrame
    update: DataFrame
    delete: DataFrame

    def __init__(self, create: DataFrame, update: DataFrame, delete: DataFrame):
        self.create = create
        self.update = update
        self.delete = delete

    def __str__(self):
        return f"Create: {self.create.height}, Update: {self.update.height}, Delete: {self.delete.height}"

    def __repr__(self):
        return str(self)



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator

from ...errors import (
    ConfigValidationError,
    ConfigValueError,
    RequiredDestinationFieldError,
)
from ..interface import SourceSync, SourceSyncArguments

polar_to_xml_type_map = {
    "Float32": "Float32",
    "Float64": "Float64",
    "Int8": "Int8",
    "Int16": "Int16",
    "Int32": "Int32",
    "Int64": "Int64",
    "Int128": "Int128",
    "UInt8": "UInt8",
    "UInt16": "UInt16",
    "UInt32": "UInt32",
    "UInt64": "UInt64",
    "String": "str",
    "str": "str",
    "Utf8": "str",
    "Boolean": "bool",
}


class XmlSyncArguments(SourceSyncArguments):
    xpath: Annotated[str, Field(default="./*")]
    unique_columns: Annotated[tuple[str, ...], Field(default=tuple([]))]
    root_name: Annotated[str, Field(default="root")]
    row_name: Annotated[str, Field(default="row")]
    types: Annotated[Optional[dict[str, str]], Field(default=None)]
    xsl: Annotated[Optional[str], Field(default=None)]
    type: Annotated[Literal["Xml"], Field(default="Xml")] = "Xml"

    @model_validator(mode="after")
    def validate_types(cls, arguments: "XmlSyncArguments"):
        errors: list[Exception] = []
        if arguments.types is not None:
            for key, value in arguments.types.items():
                if value not in polar_to_xml_type_map:
                    errors.append(
                        ConfigValueError(
                            f"types.{key}",
                            f"Type '{value}' is not supported. Supported types are {polar_to_xml_type_map.keys()}.",
                            arguments.location,
                        )
                    )
        return arguments

    @model_validator(mode="after")
    def validate_as_destination(cls, instance: "XmlSyncArguments"):
        if instance.direction == "source":
            return instance
        errors: list[Exception] = []
        if not instance.unique_columns:
            errors.append(
                RequiredDestinationFieldError("unique_columns", instance.location)
            )

        if errors:
            raise ConfigValidationError(errors)

        return instance


class XmlSync(SourceSync):
    """The configuration class used for XML file sources"""

    xpath: str = ""
    unique_columns: tuple[str] = ()
    root_name: str = ""
    row_name: str = ""
    types: dict = None
    xsl: Optional[str] = None
    parser: Literal["etree", "lxml"] = "etree"

    def __init__(self, arguments: XmlSyncArguments):
        super().__init__(arguments)
        self.xpath = arguments.xpath
        self.unique_columns = arguments.unique_columns
        self.root_name = arguments.root_name
        self.row_name = arguments.row_name

        if arguments.xsl is not None:
            self.xsl = arguments.xsl
            self.parser = "lxml"

        if arguments.types is not None:
            self.types = {}
            for key, value in arguments.types.items():
                self.types[key] = polar_to_xml_type_map[value]



import polars as pl
from .interface import TransformerInterface, register_transformer_class

NAN_SUPPORTED_TYPES = [pl.Decimal, pl.Float32, pl.Float64, pl.Int8,
                       pl.Int16, pl.Int32, pl.Int64, pl.UInt8, pl.UInt16, pl.UInt32, pl.UInt64]


@register_transformer_class("int")
class IntegerTransformer(TransformerInterface):
    @staticmethod
    def divide(data: pl.DataFrame, inField: str, outField: str, factor: int, inType: str = "Int64", outType: str = "Int32") -> pl.DataFrame:
        """Divide the numbers in a given column with the given factor to a rounded integer

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            outField (str): The field to put the output into
            factor (int): The field to divide by

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        inType = getattr(pl, inType)
        outType = getattr(pl, outType)

        data = data.with_columns(
            (data[inField].cast(inType) / factor).round(0).cast(outType).alias(outField))

        return data

    @staticmethod
    def fillna(data: pl.DataFrame, inField: str, outField: str = None, value: int = 0) -> pl.DataFrame:
        """Fill the missing NaN and null values in a given column with the given value

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            outField (str): The field to put the output into
            value (int): The value to fill

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        data = data.with_columns(data[inField].fill_null(
            value).alias(outField or inField))

        if data[inField].dtype in NAN_SUPPORTED_TYPES:
            data = data.with_columns(data[inField].fill_nan(
                value).alias(outField or inField))

        return data

    @staticmethod
    def to_int64(data: pl.DataFrame, inField: str, outField: str = None) -> pl.DataFrame:
        """Convert the numbers in a given column to Int64

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            outField (str): The field to put the output into

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        old_series = data[inField]
        new_series = old_series.cast(pl.Int64)
        series_with_alias = new_series.alias(outField or inField)
        data = data.with_columns(series_with_alias)

        return data



class ConfigValidationError(Exception):
    """An error for when a model fails to validate, can contain multiple errors"""

    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return "".join(str(error) for error in self.errors)


class ConfigValueError(Exception):
    """An error for when a field fails to validate"""

    message: str
    location: tuple[str]
    further_information_url: str = ""

    def __init__(self, field: str, message: str, location: tuple[str]):
        self.message = message
        self.location = location + (field,)

    def create_further_information_string(self):
        if self.further_information_url:
            return f"\n{'    '}For further information visit: {self.further_information_url}"

        return "\n    No further information available for this error type"

    def __str__(self):
        location = ".".join(self.location)
        further_information = self.create_further_information_string()
        return f"\n{location}\n{'  '}{self.message}{further_information}"


class InvalidDependencyError(ConfigValueError):
    def __init__(self, field: str, details: str, location: tuple[str]):
        super().__init__(
            field,
            f"Field '{field}' results in an invalid dependency: {details}",
            location,
        )

class MissingDependencyError(ConfigValueError):
    def __init__(self, field: str, details: str, location: tuple[str]):
        super().__init__(
            field,
            f"Field '{field}' results in a missing dependency: {details}",
            location,
        )

class RequiredDestinationFieldError(ConfigValueError):
    def __init__(self, field: str, destination_location: tuple[str]):
        super().__init__(
            field,
            f"Field '{field}' is required when used as destination",
            destination_location,
        )


class RequiredDestinationFieldError(ConfigValueError):
    def __init__(self, field: str, destination_location: tuple[str]):
        super().__init__(
            field,
            f"Field '{field}' is required when used as destination",
            destination_location,
        )


class RequiredSourceFieldError(ConfigValueError):
    def __init__(self, field: str, source_location: tuple[str]):
        super().__init__(
            field, f"Field '{field}' is required when used as source", source_location
        )


class ForbiddenSourceFieldError(ConfigValueError):
    def __init__(self, field: str, source_location: tuple[str]):
        super().__init__(
            field, f"Field '{field}' is forbidden when used as source", source_location
        )




from polars import DataFrame

from ..interface import SourceInterface
from ..registrated_source import register_source
from .faker_config import FakerConfig, FakerConfigArguments


@register_source("Faker")
class FakerSource(SourceInterface):
    ConfigArgumentsClass = FakerConfigArguments
    ConfigClass = FakerConfig

    """ A source for faker data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> DataFrame:
        return self.connection_settings.data

    def insert(self, data: DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)



import json

import polars as pl
import requests
import requests.adapters
import urllib3
from alive_progress import alive_bar

from ...config.polar_types import PolarTypeOverridesParameters
from ..interface import SourceInterface
from ..registrated_source import register_source
from ..request_threader import RequestThreader
from .itop_config import ItopConfig, ItopConfigArguments
from .itop_sync import ItopSync, ItopSyncArguments

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

BULK_CUTOFF = 300


@register_source("Itop")
class ItopSource(SourceInterface):
    ConfigArgumentsClass = ItopConfigArguments
    ConfigClass = ItopConfig
    SyncArgumentsClass = ItopSyncArguments
    SyncClass = ItopSync

    source_configuration: ItopSync = None
    connection_settings: ItopConfig = None
    auth_data: dict = None
    soft_deleted_items: pl.DataFrame = None

    """ A source for Combodo iTop Data data """

    def soft_delete_active(self) -> bool:
        """Check if the soft_delete option is active

        Returns:
            bool: True/False
        """
        if self.source_configuration.soft_delete is None:
            return False

        return self.source_configuration.soft_delete.enabled

    def _configure(self):
        self.auth_data = {
            "auth_user": self.connection_settings.username,
            "auth_pwd": self.connection_settings.password,
        }

        self.url = f"https://{self.connection_settings.host}/webservices/rest.php?version=1.3&login_mode=form"

        # Test Connection
        files = self._create_body(
            "core/check_credentials",
            {
                "user": self.connection_settings.username,
                "password": self.connection_settings.password,
            },
        )

        res = requests.post(
            self.url,
            files=files,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except AssertionError:
            raise Exception(
                f"Invalid response received from iTop Server: {res_json['code']}"
            )
        except requests.exceptions.JSONDecodeError:
            raise Exception(
                "Invalid response received from iTop Server, "
                "check your authentication settings: " + res.text
            )

    def _create_body(self, operation: str, params: dict = None):
        data = params or {}

        data["operation"] = operation

        return {"json_data": json.dumps(data)}

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        """
        In:
            Object: The type of object to retrieve from iTop
            OQL: OQL Query
            Output Fields: List of fields to return
        """
        oql = self.source_configuration.oql_key
        soft_deleted_oql = None
        if self.soft_delete_active():
            soft_delete = self.source_configuration.soft_delete
            if f"{soft_delete.field.lower()}" in oql.lower():
                raise Exception(
                    "Soft delete where clause is managed by beetl please remove it from the OQL."
                )

            combining_word = "WHERE" if "WHERE" not in oql.upper() else "AND"
            oql += (
                f" {combining_word} {soft_delete.field} "
                + f"= '{soft_delete.active_value}'"
            )

            soft_deleted_oql = self.source_configuration.oql_key + (
                f" {combining_word} {soft_delete.field} "
                + f"= '{soft_delete.inactive_value}'"
            )

        all_colums = (
            self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
            + self.source_configuration.link_columns
        )

        if soft_deleted_oql:
            self.soft_deleted_items = self._run_oql_query(
                self.source_configuration.datamodel,
                soft_deleted_oql,
                self.source_configuration.unique_columns,
            )

        return self._run_oql_query(
            self.source_configuration.datamodel,
            oql,
            all_colums,
            self.source_configuration.type_overrides,
        )

    def _run_oql_query(
        self,
        data_model: str,
        oql: str,
        all_colums: tuple[str],
        type_overrides: PolarTypeOverridesParameters = {},
    ) -> pl.DataFrame:
        files = self._create_body(
            "core/get",
            {
                "class": data_model,
                "key": oql,
                "output_fields": ",".join(all_colums),
            },
        )

        res = requests.post(
            self.url,
            files=files,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok or res.json()["code"] != 0:
            raise Exception(
                f"Error retrieving {self.source_configuration.datamodel}"
                f"data from iTop: {res.text}"
            )

        re_obj = []

        try:
            res_json = res.json()
            objects = res_json.get("objects", None)
            if objects is None:
                return pl.DataFrame()

            for _, item in res_json.get("objects", {}).items():
                if "id" in all_colums:
                    re_obj.append({"id": item["key"], **item["fields"]})
                elif "key" in all_colums:
                    re_obj.append({"key": item["key"], **item["fields"]})
                else:
                    re_obj.append(item["fields"])

            return pl.DataFrame(re_obj, schema_overrides=type_overrides)

        except Exception as e:
            raise Exception(
                "An error occured while trying to decode the response", e
            ) from e

    def create_item(self, type: str, data: dict, comment: str = ""):
        self.set_foreign_key_none_values_to_itop_unassigned(data)

        body = self._create_body(
            "core/create",
            {
                "comment": comment,
                "class": type,
                "output_fields": "*",
                "fields": {
                    x: y
                    for x, y in data.items()
                    if y is not None and x not in self.source_configuration.link_columns
                },
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not created: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Created", e, res.text, body)
            print(res.text + str(e))
            return False

        return True

    def update_item(self, type: str, data: dict, comment: str = ""):
        identifiers = {}

        for column_name in tuple(data.keys()):
            if column_name in self.source_configuration.unique_columns:
                identifiers[column_name] = data.pop(column_name)

        where_clause_conditions = " AND ".join(
            [f"{key} = '{value}'" for key, value in identifiers.items()]
        )

        self.set_foreign_key_none_values_to_itop_unassigned(data)

        body = self._create_body(
            "core/update",
            {
                "comment": comment,
                "class": type,
                "key": f"SELECT {type} WHERE {where_clause_conditions}",
                "fields": {
                    key: value for key, value in data.items() if value is not None
                },
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not updated: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Updated", e, res.text, body)
            return False

        return True

    def set_foreign_key_none_values_to_itop_unassigned(self, data):
        for key, value in data.items():
            if key in self.source_configuration.foreign_key_columns:
                if value is None:
                    data[key] = 0

    def delete_item(self, type: str, data: dict, comment: str = ""):
        identifiers = {}

        for key in tuple(data.keys()):
            if key in self.source_configuration.unique_columns:
                identifiers[key] = data.pop(key)

        where_clause_conditions = " AND ".join(
            [f"{key} = '{value}'" for key, value in identifiers.items()]
        )

        body = self._create_body(
            "core/delete",
            {
                "comment": comment,
                "class": type,
                "key": f"SELECT {type} WHERE {where_clause_conditions}",
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not deleted: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Deleted", e, res.text, body)
            return False

        return True

    def _insert(self, data: pl.DataFrame):
        insertData = data.clone()
        insert_cols = tuple(
            self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
        )

        for column in data.columns:
            if column not in insert_cols:
                insertData.drop_in_place(column)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": "Created via API Sync",
            }
            for x in insertData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(50) as needle:
                    for _ in needle.submitAndWait(self.create_item, iters):
                        pr_bar()
            else:
                for item in iters:
                    self.create_item(**item)
                    pr_bar()

    def insert(self, data: pl.DataFrame):
        if (
            self.soft_delete_active()
            and self.soft_deleted_items is not None
            and len(self.soft_deleted_items) > 0
        ):
            # When soft delete is active we need to update the soft deleted items that are being reinserted.
            # We do this to avoid creating duplicates in iTop.
            self.update(
                data.join(
                    self.soft_deleted_items,
                    on=self.source_configuration.unique_columns,
                    how="inner",
                )
            )

            self._insert(
                data.join(
                    self.soft_deleted_items,
                    on=self.source_configuration.unique_columns,
                    how="anti",
                )
            )
        else:
            self._insert(data)

        return len(data)

    def update(self, data: pl.DataFrame):
        updateData = data.clone()

        update_cols = tuple(
            column_name
            for column_name in self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
            if column_name not in (self.source_configuration.skip_columns or [])
        )

        for column in data.columns:
            if column not in update_cols:
                updateData.drop_in_place(column)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": "Updated via API Sync",
            }
            for x in updateData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(50) as needle:
                    for _ in needle.submitAndWait(self.update_item, iters):
                        pr_bar()
            else:
                for item in iters:
                    self.update_item(**item)
                    pr_bar()

        return len(data)

    def delete(self, data: pl.DataFrame):
        deleteData = data.clone()
        deleteFunc = self.delete_item
        deleteMessage = "Deleted via API Sync"

        if self.soft_delete_active():
            soft_delete = self.source_configuration.soft_delete

            deleteData = deleteData.with_columns(
                pl.Series(
                    soft_delete.field,
                    [soft_delete.inactive_value] * len(deleteData),
                )
            )
            deleteFunc = self.update_item
            deleteMessage = "Soft deletion via API Sync"

        update_cols = set(
            column_name
            for column_name in self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
            if column_name not in (self.source_configuration.skip_columns or [])
        )

        deleteData = deleteData.select(update_cols)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": deleteMessage,
            }
            for x in deleteData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(100) as needle:
                    for _ in needle.submitAndWait(deleteFunc, iters):
                        pr_bar()
            else:
                for item in iters:
                    deleteFunc(**item)
                    pr_bar()

        return len(data)



import os

# re-export
from .csv import CsvConfig, CsvConfigArguments, CsvSource
from .faker import FakerConfig, FakerConfigArguments, FakerSource
from .interface import (
    CASTABLE,
    SourceConfig,
    SourceConfigArguments,
    SourceConnectionArguments,
    SourceSync,
    SourceSyncArguments,
)
from .itop import ItopConfig, ItopConfigArguments, ItopSource, ItopSyncArguments
from .mongodb import (
    MongodbConfig,
    MongodbConfigArguments,
    MongodbSource,
    MongodbSync,
    MongodbSyncArguments,
)
from .mysql import MysqlConfig, MysqlConfigArguments, MysqlSource, MysqlSyncArguments
from .postgresql import (
    PostgresConfig,
    PostgresConfigArguments,
    PostgresSource,
    PostgresSyncArguments,
)
from .registrated_source import Sources, register_source
from .request_threader import RequestThreader
from .rest import RestConfig, RestConfigArguments, RestSource, RestSyncArguments
from .sqlserver import (
    SqlserverConfig,
    SqlserverConfigArguments,
    SqlserverSource,
    SqlserverSync,
    SqlserverSyncArguments,
)
from .static import StaticConfig, StaticConfigArguments, StaticSource
from .excel import ExcelConfig, ExcelConfigArguments, ExcelSource

from .xml import XmlConfig, XmlConfigArguments, XmlSource, XmlSync, XmlSyncArguments


cDir = os.path.dirname(__file__)
for file in os.listdir(cDir):
    if file.endswith(".py") and not file.startswith("__") and file != "interface.py":
        __import__(f"{__name__}.{file[:-3]}", fromlist=["*"])



import yaml
import json
from typing import Any, List, Dict, Literal
from dataclasses import dataclass
import polars as pl
from ..sources.interface import (
    SourceSync,
    SourceConfig,
)
from ..transformers.interface import TransformerConfiguration


class SourceSettings:
    """The configuration class used for datasources"""

    name: str
    source_type: str
    connection: SourceConfig
    config: SourceSync

    def __init__(
        self, name: str, source_type: str, connection: dict, config: dict
    ) -> None:
        self.name, self.source_type = name, source_type
        source_class = __import__(f"beetl.sources.{source_type.lower()}")

        self.connection = getattr(
            source_class, f"{source_type}SourceConnectionSettings"
        )(**connection)

        self.config = getattr(
            source_class, f"{source_type}SourceConfig")(**config)


@dataclass
class ChangeDetectionConfig:
    threshold: int = -1
    stopSync: bool = False
    webhook: str = None
    webhookHeaders: Dict[str, str] = None
    webhookPayload: Dict[str, str] = None


@dataclass
class ComparisonColumn:
    name: str
    type: pl.DataType
    unique: bool = False

    def __init__(self, name: str, type: Any, unique: bool = False) -> None:
        self.name, self.unique = name, unique
        self.type = getattr(pl, type)


@dataclass
class SyncConfiguration:
    """The configuration for a single sync between two sources"""

    source: SourceSettings
    sourceConfig: SourceSync
    destination: SourceSettings
    destinationConfig: SourceSync
    comparisonColumns: List[ComparisonColumn]
    name: str = ""
    changeDetection: ChangeDetectionConfig = None
    sourceTransformers: TransformerConfiguration = None
    destinationTransformers: TransformerConfiguration = None
    insertionTransformers: List[TransformerConfiguration] = None
    deletionTransformers: List[TransformerConfiguration] = None

    def __post_init__(self) -> None:
        self.source.config = self.sourceConfig
        self.destination.config = self.destinationConfig


class BeetlConfig:
    """The configuration class for BeETL"""

    version: Literal["V1"]
    sources: Dict[str, SourceSettings]
    sync_list: List[SyncConfiguration]

    def __init__(self, config: dict) -> None:
        # Set version to keep track of changes
        # in an effort to keep backwards compatibility

        # Get the class corresponding to the configuration version
        module_root = str.join(".", self.__module__.split(".")[:-1])
        module = __import__(module_root, fromlist=[""])
        config_class = getattr(
            module, f'BeetlConfig{config.get("configVersion", "V1")}'
        )

        # Transform this class into the versioned configuration class
        self.__class__ = config_class
        self.__dict__ = config_class(config).__dict__
        self.version = config.get("configVersion", "V1")

    @classmethod
    def from_yaml(cls, yamlData: str) -> "BeetlConfig":
        """Import configuration from a YAML string

        Args:
            yaml (str): The YAML string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = yaml.safe_load(yamlData)
        return cls(config)

    @classmethod
    def from_yaml_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = yaml.safe_load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e

    @classmethod
    def from_json(cls, jsonData: str) -> "BeetlConfig":
        """Import configuration from a JSON string

        Args:
            jsonData (str): The JSON string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = json.loads(jsonData)
        return cls(config)

    @classmethod
    def from_json_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = json.load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class CsvConnectionArguments(SourceConnectionArguments):
    path: Annotated[str, Field(min_length=1)]
    encoding: Annotated[str, Field(default="utf-8")]


class CsvConfigArguments(SourceConfigArguments):

    type: Annotated[Literal["Csv"], Field(default="Csv")] = "Csv"
    connection: CsvConnectionArguments


class CsvConfig(SourceConfig):
    """The connection configuration class used for static sources"""

    path: str
    encoding: str

    def __init__(self, arguments: CsvConfigArguments):
        super().__init__(arguments)
        self.path = arguments.connection.path
        self.encoding = arguments.connection.encoding



from typing import Annotated, Literal

from pydantic import Field

from ...validation import ValidationBaseModel


class SourceSyncArguments(ValidationBaseModel):
    """Class representation of the source configuration settings in the beetl config. Used to validate the source configuration settings using pydantic."""

    name: Annotated[str, Field(min_length=1)]
    type: Annotated[Literal["Interface"], Field(default="Interface")] = "Interface"
    direction: Annotated[Literal["source", "destination"], Field(min_length=1)]


class SourceSync:
    """The configuration class used for data sources, abstract"""

    def __init__(cls, arguments: SourceSyncArguments):
        pass



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator
from pyodbc import DatabaseError, drivers

from ...errors import ConfigValidationError, ConfigValueError, InvalidDependencyError, MissingDependencyError
from ...validation import ValidationBaseModel
from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class SqlserverSettingsArguments(ValidationBaseModel):
    connection_string: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Required if individual connection settings are not provided",
        ),
    ]
    username: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    password: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    host: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    port: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    database: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]

    @model_validator(mode="after")
    def validate_connection_settings(cls, instance: "SqlserverConnectionArguments"):
        errors = []
        
        sql_drivers = [driver for driver in drivers() if "SQL Server" in driver]
        if len(sql_drivers) == 0:
           errors.append(MissingDependencyError(
               "connection_string",
               "No ODBC drivers were found for SQL Server/pyodbc. Please install the Microsoft ODBC Driver for SQL Server",
               instance.location
           ))
        
        if len(sql_drivers) == 1 and sql_drivers[0] == "SQL Server":
            errors.append(InvalidDependencyError(
                "connection_string",
                "Your system has an old version of the ODBC driver for SQL Server."
                "Please install a more recent version of the driver (Microsoft ODBC Driver 17 for SQL Server or above)",
                instance.location
            ))
        
        if instance.connection_string:
            if "driver=SQL+Server" in instance.connection_string or "driver=SQL Server" in instance.connection_string:
                errstring = "You need to download a more recent version of the driver (Microsoft ODBC Driver 17 for SQL Server or above)"
                found = [ driver for driver in drivers() if  driver.startswith("ODBC") ]
                if len(found) > 0:
                    errstring = f"You have the following ODBC drivers installed, please use one of them or download a more recent version: {found.join(', ')}"
                    
                errors.append(InvalidDependencyError(
                    "connection_string",
                    "The connection string contains the driver 'SQL Server' or 'SQL+Server'. " + errstring,
                    instance.location
                ))
            
            if len(sql_drivers) == 1 and sql_drivers[0] == "SQL Server":
                errors.append(
                    ConfigValueError(
                        "connection_string"
                    )
                )
            
            return instance

        required_fields = ["username", "password", "host", "port", "database"]
        for field in required_fields:
            if not getattr(instance, field):
                errors.append(
                    ConfigValueError(
                        field,
                        f"Field '{field}' is required when 'connection_string' is not provided",
                        instance.location,
                    )
                )
        
        if errors:
            raise ConfigValidationError(errors)
        
        return instance


class SqlserverConnectionArguments(SourceConnectionArguments):
    settings: SqlserverSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, arguments: "SqlserverConnectionArguments"):
        cls.propagate_location("settings", arguments)
        return arguments


class SqlserverConfigArguments(SourceConfigArguments):
    connection: SqlserverConnectionArguments
    type: Annotated[Literal["Sqlserver"], Field(default="Sqlserver")] = "Sqlserver"


class SqlserverConfig(SourceConfig):
    """The connection configuration class used for SQLServer sources"""

    connection_string: str

    def append_sqlserver_driver(self, connection_string: str):
        if "driver" in connection_string:
            return connection_string

        sql_drivers = [driver for driver in drivers() if "SQL Server" in driver]
        no_drivers_available = len(sql_drivers) == 0
        if no_drivers_available:
            raise DatabaseError(
                "No ODBC drivers found for SQL Server/pyodbc. "
                "Please make sure at least one is installed"
            )

        print(
            f"\nFound driver '{sql_drivers[0]}', beetl is automatically appending it to the connection string for you. If you want to use a different driver, please add it manually."
        )
        join_char = "&" if "?" in connection_string else "?"
        return f"{connection_string}{join_char}driver={sql_drivers[0]}"

    def __init__(self, arguments: SqlserverConfigArguments):
        super().__init__(arguments)
        if arguments.connection.settings.connection_string:
            self.connection_string = self.append_sqlserver_driver(
                arguments.connection.settings.connection_string
            )
            return

        composed_connection_string = "mssql://"
        f"{arguments.connection.settings.username}:{arguments.connection.settings.password}"
        f"@{arguments.connection.settings.host}:{arguments.connection.settings.port}/{arguments.connection.settings.database}"

        self.connection_string = self.append_sqlserver_driver(
            composed_connection_string
        )



from dataclasses import dataclass

import polars as pl
from polars import DataType


@dataclass
class ColumnDefinition:
    """The definition of a column in a dataset.
    Name: Name of column
    Type: Polars data type
    Unique: Whether it is unique
    Skip Update: Whether to skip updating this field when inserting/updating source
    Custom Options: Custom options for some providers
    """

    name: str
    type: DataType
    unique: bool = False
    skip_update: bool = False
    custom_options: dict = None

    def __post_init__(self) -> None:
        """Get the actual data type from Polars"""
        self.type = getattr(pl, self.type)



import os

import pandas as pd
import polars as pl

from ..interface import SourceInterface
from ..registrated_source import register_source
from .xml_config import XmlConfig, XmlConfigArguments
from .xml_sync import XmlSync, XmlSyncArguments


@register_source("Xml")
class XmlSource(SourceInterface):
    """A source for reading and writing XML files"""

    ConfigArgumentsClass = XmlConfigArguments
    ConfigClass = XmlConfig
    SyncArgumentsClass = XmlSyncArguments
    SyncClass = XmlSync

    connection_settings: XmlConfig = None
    source_configuration: XmlSync = None
    mutation_data: pl.DataFrame = pl.DataFrame()

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        """Flushes the mutated dataframe to the xml file"""
        self.flush_data_to_file(self.mutation_data)

    def _query(self, params=None) -> pl.DataFrame:
        """Reads the XML file and returns a polars DataFrame"""

        file_does_not_exist = not os.path.exists(self.connection_settings.path)
        if file_does_not_exist:
            return pl.DataFrame()

        file_is_empty = os.path.getsize(self.connection_settings.path) == 0
        if file_is_empty:
            return pl.DataFrame()

        data = pl.from_pandas(
            pd.read_xml(
                self.connection_settings.path,
                encoding=self.connection_settings.encoding,
                parser=self.source_configuration.parser,
                xpath=self.source_configuration.xpath,
                dtype=self.source_configuration.types,
                stylesheet=self.source_configuration.xsl,
            )
        )

        return data

    def insert(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.update(
            data, on=unique_columns, how="full", include_nulls=True
        )
        self.save_data_after_mutation(updated)
        return len(data)

    def update(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.update(
            data, on=unique_columns, how="left", include_nulls=True
        )
        self.save_data_after_mutation(updated)
        return len(data)

    def delete(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.join(data, on=unique_columns, how="anti")
        self.save_data_after_mutation(updated)
        return len(data)

    def load_data_for_mutation(self, data_schema: pl.DataFrame) -> pl.DataFrame:
        if self.mutation_data.height > 0:
            return self.mutation_data

        mutation_data = self._query()
        self.mutation_data = self.append_missing_columns(data_schema, mutation_data)

        return self.mutation_data

    def save_data_after_mutation(self, data: pl.DataFrame) -> None:
        self.mutation_data = data

    def flush_data_to_file(self, data: pl.DataFrame) -> None:
        data_frame_is_missing_columns = data.width == 0
        if data_frame_is_missing_columns:
            return

        if not os.path.exists(self.connection_settings.path):
            with open(self.connection_settings.path, "w") as file:
                file.flush()

        data.to_pandas().to_xml(
            self.connection_settings.path,
            encoding=self.connection_settings.encoding,
            root_name=self.source_configuration.root_name,
            row_name=self.source_configuration.row_name,
            index=False,
            parser="etree",
        )

    def get_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "Unique columns must be defined in the configuration when used as a destination"
            )

        return self.source_configuration.unique_columns

    def append_missing_columns(
        self, source: pl.DataFrame, destination: pl.DataFrame
    ) -> pl.DataFrame:
        columns_to_append = [
            column for column in source.columns if column not in destination.columns
        ]
        for column in columns_to_append:
            destination = destination.with_columns(
                pl.Series(
                    column, [None] * destination.height, dtype=source[column].dtype
                )
            )

        return destination



from typing import Annotated, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class IntegerTransformerSchema:
    class Divide(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[str, Field(min_length=1)]
            factor: int
            inType: Annotated[str, Field(default="Int64")]
            outType: Annotated[str, Field(default="Int32")]

        transformer: Literal["int.divide"]
        config: Config

    class Fillna(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            value: Annotated[int, Field(default=0)]

        transformer: Literal["int.fillna"]
        config: Config

    class ToInt64(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]

        transformer: Literal["int.to_int64"]
        config: Config


IntegerTransformerSchemas = Union[
    IntegerTransformerSchema.Divide,
    IntegerTransformerSchema.Fillna,
    IntegerTransformerSchema.ToInt64,
]



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator

from ..interface import SourceConfig, SourceConfigArguments


class ItopConfigArguments(SourceConfigArguments):
    class ItopConnectionArguments(BaseModel):
        model_config = ConfigDict(extra="forbid")

        host: Annotated[str, Field(min_length=1)]
        username: Annotated[str, Field(min_length=1)]
        password: Annotated[str, Field(min_length=1)]
        verify_ssl: Annotated[str, Field(default="true")]

        @model_validator(mode="before")
        def transform_input(cls, values):
            settings: dict[str, str] = values.get("settings", {})
            transformed_values: dict[str, str] = {}
            transformed_values["host"] = settings.get("host", None)
            transformed_values["username"] = settings.get("username", None)
            transformed_values["password"] = settings.get("password", None)
            transformed_values["verify_ssl"] = settings.get("verify_ssl", "true")

            if type(transformed_values["verify_ssl"]) is bool:
                transformed_values["verify_ssl"] = str(
                    transformed_values["verify_ssl"]
                ).lower()
            return transformed_values

    type: Annotated[Literal["Itop"], Field(default="Itop")] = "Itop"
    connection: ItopConnectionArguments


class ItopConfig(SourceConfig):
    """The connection configuration class used for iTop sources"""

    host: str
    username: str
    password: str
    verify_ssl: bool

    def __init__(self, arguments: ItopConfigArguments):
        super().__init__(arguments)
        self.host = arguments.connection.host
        self.username = arguments.connection.username
        self.password = arguments.connection.password
        self.verify_ssl = arguments.connection.verify_ssl == "true"



import polars as pl


class Difftool:

    @staticmethod
    def diff_update(
        update: pl.DataFrame,
        destination: pl.DataFrame,
        column_names_props: list[str],
        key_columns: list[str],
        ignore_dates=False,
    ) -> tuple[list[pl.DataFrame], set[str]]:
        """
        This function compares two dataframes and returns a list of dataframes with the differences and a set with the keys that have differences across the whole dataframe.
        It is not very efficient, but very helpful when doing data validation on a sync that is going to affect an already existing dataset.

        Args:
            update (pl.DataFrame): The dataframe returned by the beetl comparer containing the items going to be updated.
            destination (pl.DataFrame): The destination dataframe after being transformed using the source transformers.
            column_names_props (list[str]): The list of column names that are going to be compared.
            key_columns (list[str]): The list of columns that are going to be used as keys.
            ignore_dates (bool, optional): If true, it will ignore columns that have "date" in their name. Defaults to False.

        Returns:
            tuple[list[pl.DataFrame], set[str]]: A tuple containing a list of dataframes with the differences and a set with the keys that have differences across the whole dataframe
        """
        column_names = [*column_names_props]
        if ignore_dates:
            column_names = [name for name in column_names if "date" not in name.lower()]

        calcultated_dataframe_diffs = []
        keys_that_differ_across_whole_dataset = set()

        for row_index in range(update.height):
            current_row_keys_dataframe = update.select(key_columns)[row_index]
            current_update_dataframe_row = update.join(
                current_row_keys_dataframe, on=key_columns
            ).select(column_names)
            current_destination_dataframe_row = destination.join(
                current_row_keys_dataframe, on=key_columns
            ).select(column_names)

            multiple_rows_found_using_keys = (
                current_update_dataframe_row.height > 1
                or current_destination_dataframe_row.height > 1
            )
            if multiple_rows_found_using_keys:
                raise Exception(
                    f"Multiple rows were found when using the keys {key_columns} for row index {row_index} in the update dataset. This is not allowed since it will cause beetl to miss or overwrite data that should't be overwritten."
                )

            current_row_comparison_dataframe = (
                current_update_dataframe_row == current_destination_dataframe_row
            )
            current_row_comparison_dict = current_row_comparison_dataframe.to_dict()
            keys_that_differ_in_current_row = []
            for key, value in current_row_comparison_dict.items():
                if not value[0]:
                    keys_that_differ_in_current_row.append(key)

            key_values_for_current_row = list(
                map(str, list(current_row_keys_dataframe.to_dicts()[0].values()))
            )

            current_diff_dataframe = pl.DataFrame().with_columns(
                pl.Series("dataset", ["update", "destination"]),
                pl.Series(
                    "keys", [key_values_for_current_row, key_values_for_current_row]
                ),
            )
            current_row_has_diff = False
            for key in keys_that_differ_in_current_row:
                value1 = current_update_dataframe_row[key][0]
                value2 = current_destination_dataframe_row[key][0]
                if value1 is None and value2 is None:
                    continue
                if value1 == value2:
                    continue
                series = pl.Series(key, [value1, value2])
                current_diff_dataframe = current_diff_dataframe.with_columns(series)
                current_row_has_diff = True

            if not current_row_has_diff:
                continue

            calcultated_dataframe_diffs.append(current_diff_dataframe)
            keys_that_differ_across_whole_dataset.update(
                keys_that_differ_in_current_row
            )

        return (calcultated_dataframe_diffs, keys_that_differ_across_whole_dataset)



from typing import Annotated, Any, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class ItopTransformerSchema:
    class Orgcode(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inFields: Annotated[list[str], Field(min_length=1)]
            outField: Annotated[str, Field(min_length=1)]
            toplevel: Annotated[Optional[str], Field(default=None)]

        transformer: Literal["itop.orgcode"]
        config: Config

    class Orgtree(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            treeFields: Annotated[list[str], Field(min_length=1)]
            toplevel: Annotated[str, Field(min_length=1)]
            name_field: Annotated[str, Field(default="name")]
            path_field: Annotated[str, Field(default="orgpath")]
            code_field: Annotated[str, Field(default="code")]
            parent_field: Annotated[str, Field(default="parent_code")]

        transformer: Literal["itop.orgtree"]
        config: Config

    class Relations(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            class FieldRelation(TransformerConfigBase):
                source_field: Annotated[str, Field(min_length=1)]
                source_comparison_field: Annotated[str, Field(min_length=1)]
                foreign_class_type: Annotated[str, Field(min_length=1)]
                foreign_comparison_field: Annotated[str, Field(min_length=1)]
                use_like_operator: Annotated[bool, Field(default=False)]

            field_relations: Annotated[list[FieldRelation], Field(min_length=1)]

        transformer: Literal["itop.relations"]
        config: Config


ItopTransformerSchemas = Union[
    ItopTransformerSchema.Orgcode,
    ItopTransformerSchema.Orgtree,
    ItopTransformerSchema.Relations,
]



from .interface import SourceInterface


class RegistratedSource:
    name: str
    cls: SourceInterface

    def __init__(self, name: str, cls: SourceInterface):
        self.name = name
        self.cls = cls


def register_source(name: str):
    def wrapper(cls: type):
        Sources.sources[name] = RegistratedSource(name, cls)

        return cls

    return wrapper


class Sources:
    sources: dict[str, RegistratedSource] = {}



from typing import Annotated, Any, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class Mapping(TransformerConfigBase):
    # Needs an alias since from is a protected keyword
    from_: Annotated[str, Field(min_length=1, alias="from")]
    to: Annotated[str, Field(min_length=1)]


class FramesTransformerSchema:
    class Filter(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            filter: Annotated[dict[str, Any], Field(min_items=1)]
            reverse: Annotated[Optional[bool], Field(default=False)]

        transformer: Literal["frames.filter"]
        config: Config

    class Conditional(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            conditionField: Annotated[str, Field(min_length=1)]
            ifTrue: str
            ifFalse: str
            targetField: Annotated[str, Field(default="")]

        transformer: Literal["frames.conditional"]
        config: Config

    class RenameColumns(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            columns: Annotated[
                Union[list[Mapping], dict[str, str]], Field(min_length=1)
            ]

        transformer: Literal["frames.rename_columns"]
        config: Config

    class CopyColumns(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            columns: Annotated[list[Mapping], Field(min_length=1)]

        transformer: Literal["frames.copy_columns"]
        config: Config

    class DropColumns(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            columns: Annotated[list[str], Field(min_length=1)]

        transformer: Literal["frames.drop_columns"]
        config: Config

    class ProjectColumns(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            columns: Annotated[list[str], Field(min_length=1)]

        transformer: Literal["frames.project_columns"]
        config: Config

    class Distinct(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            columns: Annotated[list[str], Field(min_length=1)]

        transformer: Literal["frames.distinct"]
        config: Config

    class ExtractNestedRows(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            iterField: Annotated[str, Field(default="")]
            fieldMap: Annotated[dict[str, str], Field(default={})]
            colMap: Annotated[dict[str, str], Field(default={})]

        transformer: Literal["frames.extract_nested_rows"]
        config: Config


FramesTransformerSchemas = Union[
    FramesTransformerSchema.Filter,
    FramesTransformerSchema.Conditional,
    FramesTransformerSchema.RenameColumns,
    FramesTransformerSchema.CopyColumns,
    FramesTransformerSchema.DropColumns,
    FramesTransformerSchema.ProjectColumns,
    FramesTransformerSchema.Distinct,
    FramesTransformerSchema.ExtractNestedRows,
]



from typing import Annotated, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class MiscTransformerSchema:
    class SamFromDn(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]

        transformer: Literal["misc.sam_from_dn"]
        config: "MiscTransformerSchema.SamFromDn.Config"


MiscTransformerSchemas = Union[MiscTransformerSchema.SamFromDn,]



from hashlib import sha1
from typing import Any

import polars as pl

from .interface import TransformerInterface, register_transformer_class


def concat_and_sha(separator: str = "-", *args):
    retn = separator.join((x.lower() for x in args if x is not None))
    return sha1(retn.encode("utf-8")).hexdigest()


def never_none(value):
    if value is None or value == "":
        return "-"
    return value


@register_transformer_class("itop")
class ItopTransformer(TransformerInterface):
    @staticmethod
    def orgcode(data: pl.DataFrame, inFields: list, outField: str, toplevel: str):
        """Takes a number of columns, joins them and hashes the result"""

        def make_code(st):
            return concat_and_sha("-", toplevel, *st.values())

        new = data.with_columns(
            pl.struct(inFields)
            .map_elements(make_code, return_dtype=str)
            .alias(outField)
        )

        return new

    @staticmethod
    def orgtree(
        data: pl.DataFrame,
        treeFields: list,
        toplevel: str,
        name_field: str = "name",
        path_field: str = "orgpath",
        code_field: str = "code",
        parent_field: str = "parent_code",
    ):
        """Takes a number of columns and joins then together as an organization tree"""
        output = [
            {
                "status": "active",
                name_field: toplevel,
                path_field: f"Top->{toplevel}",
                code_field: concat_and_sha("-", toplevel),
                parent_field: None,
            }
        ]

        for i in range(len(treeFields)):
            for lvl in data[treeFields[: i + 1]].unique().iter_rows():
                lvls = [toplevel] + [never_none(x) for x in lvl]

                output.append(
                    {
                        "status": "active",
                        name_field: never_none(lvls[-1]),
                        path_field: "Top->" + "->".join(lvls),
                        code_field: concat_and_sha("-", *lvls),
                        parent_field: concat_and_sha("-", *lvls[:-1]),
                    }
                )

        return pl.DataFrame(output)

    @staticmethod
    def relations(
        data: pl.DataFrame, field_relations: list[dict[str, Any]]
    ) -> pl.DataFrame:
        """Insert data into iTop

        Args:
            data (pl.DataFrame): The dataFrame to modify
            connection (SourceInterface): The connection to the source
            configuration (SourceInterfaceConfiguration): The configuration for the source

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        if not field_relations:
            return data

        transformed = data.clone()
        for field_relation in field_relations:
            source_field = field_relation["source_field"]
            source_comparison_field = field_relation["source_comparison_field"]
            foreign_class_type = field_relation["foreign_class_type"]
            foreign_comparison_field = field_relation["foreign_comparison_field"]
            use_like_operator = field_relation.get("use_like_operator", False)

            if (
                not source_field
                or not source_comparison_field
                or not foreign_class_type
                or not foreign_comparison_field
            ):
                raise ValueError(
                    "One of the field_relations config values for the itop.relations transformer is missing. All fields except use_like_operator are required."
                )

            try:
                comparison_operator = "LIKE" if use_like_operator else "="

                query = (
                    f"SELECT {foreign_class_type}"
                    + f" WHERE {foreign_comparison_field} {comparison_operator} "
                )

                transformed = transformed.with_columns(
                    transformed[source_comparison_field]
                    .map_elements(
                        lambda source_comparison_field_value: (
                            query + f"'{source_comparison_field_value}'"
                            if source_comparison_field_value is not None
                            and source_comparison_field_value != ""
                            else None
                        ),
                        return_dtype=str,
                        skip_nulls=False,
                    )
                    .alias(source_field)
                )
            except pl.exceptions.ColumnNotFoundError as e:
                raise ValueError(
                    f"Error: The source_comparison_field {source_comparison_field} is not present in DataFrame"
                ) from e
            except KeyError as e:
                raise Exception("Error: The key definition is not valid") from e

        return transformed



import pandas as pd
import polars as pl
import requests
from pyarrow.lib import ArrowInvalid

from ..interface import SourceInterface
from ..registrated_source import register_source
from .rest_config import RestConfig, RestConfigArguments
from .rest_sync import RestSync, RestSyncArguments


@register_source("Rest")
class RestSource(SourceInterface):
    ConfigArgumentsClass = RestConfigArguments
    ConfigClass = RestConfig
    SyncArgumentsClass = RestSyncArguments
    SyncClass = RestSync

    source_configuration: RestSync
    connection_settings: RestConfig
    client = None
    authValidTo = None

    """ A source for MySQL data """

    def _configure(self):
        if self.client is None:
            self.client = requests.Session()
            if not self.connection_settings.authentication:
                return
            if self.connection_settings.authentication.basic:
                self.client.auth = (
                    self.connection_settings.authentication.basic_user,
                    self.connection_settings.authentication.basic_pass,
                )

            if self.connection_settings.authentication.bearer:
                self.client.headers.update(
                    {
                        "Authorization": f"{self.connection_settings.authentication.bearer_prefix} {self.connection_settings.authentication.bearer_token}"
                    }
                )

            if self.connection_settings.authentication.oauth2:
                if (
                    self.connection_settings.authentication.oauth2_settings.flow
                    == "client_credentials"
                ):
                    response = self.client.post(
                        self.connection_settings.authentication.oauth2_settings.token_url,
                        data=self.connection_settings.authentication.oauth2_settings.token_body,
                    )
                    if response.status_code != 200:
                        raise Exception("Failed to authenticate: ", response.text)

                    response_data = response.json()
                    self.client.headers.update(
                        {
                            "Authorization": f"{self.connection_settings.authentication.bearer_prefix} {response_data[self.connection_settings.authentication.oauth2_settings.token_path]}"
                        }
                    )

    def _connect(self):
        if self.authValidTo is None:
            return True

        if self.authValidTo < pd.Timestamp.now():
            return self._configure()

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        self._connect()

        lrSettings = self.source_configuration.listRequest

        request = {
            "method": lrSettings.method,
            "url": self.connection_settings.base_url + lrSettings.path,
            "headers": {"Content-Type": lrSettings.body_type},
            "verify": (not self.connection_settings.ignore_certificates),
            "params": {},
        }

        if lrSettings.return_type != "":
            request["headers"]["Accept"] = lrSettings.return_type

        if lrSettings.body_type == "application/json" and lrSettings.body is not None:
            request["json"] = lrSettings.body

        if lrSettings.query is not None:
            request["params"] = lrSettings.query

        else:
            request["data"] = lrSettings.body

        if lrSettings.pagination is not None and lrSettings.pagination.enabled:
            response_data = []
            page = lrSettings.pagination.startPage
            pageSize = lrSettings.pagination.pageSize
            total = 1

            while len(response_data) < total:
                request["params"][lrSettings.pagination.pageQuery] = page
                request["params"][lrSettings.pagination.pageSizeQuery] = pageSize

                response = self.client.request(**request)
                if response.status_code != 200:
                    if len(response_data) > 0:
                        break
                    raise Exception("Failed request")

                if lrSettings.return_type == "application/json":
                    page_response_data = response.json()
                else:
                    raise Exception("Not implemented")

                if lrSettings.pagination.totalQueryIn != "":
                    if lrSettings.pagination.totalQueryIn == "body":
                        tq = page_response_data
                        for key in lrSettings.pagination.totalQuery.split("."):
                            tq = tq[key]

                        if tq > total:
                            total = tq

                    elif lrSettings.pagination.totalQueryIn == "header":
                        total = response.headers[lrSettings.pagination.totalQuery]

                if len(lrSettings.response.items) > 0:
                    for key in lrSettings.response.items.split("."):
                        page_response_data = page_response_data[key]

                if len(response_data) == 0 and len(page_response_data) < pageSize:
                    pageSize = len(page_response_data)

                response_data += page_response_data
                page += 1

        else:
            response = self.client.request(**request)

            if response.status_code != 200:
                raise Exception("Failed request: ", response.text)

            if self.source_configuration.listRequest.return_type == "application/json":
                response_data = response.json()
            else:
                raise Exception("Not implemented")

            if (
                self.source_configuration.listRequest.response
                and self.source_configuration.listRequest.response.items
                and len(self.source_configuration.listRequest.response.items) > 0
            ):
                for key in self.source_configuration.listRequest.response.items.split(
                    "."
                ):
                    response_data = response_data[key]

            if len(response_data) == 0:
                return pl.DataFrame()

        pd_df = pd.json_normalize(response_data)

        try:
            return pl.from_pandas(pd_df)
        except ArrowInvalid:
            for col in pd_df.columns:
                if pd_df[col].dtype == "object":
                    pd_df[col] = pd_df[col].astype("string")

            return pl.from_pandas(pd_df)

    def _insert(
        self,
        data: pl.DataFrame,
    ):
        self._connect()

        request = {
            "method": self.source_configuration.createRequest.method,
            "url": self.connection_settings.base_url
            + self.source_configuration.createRequest.path,
            "headers": {
                "Content-Type": self.source_configuration.createRequest.body_type
            },
            "verify": (not self.connection_settings.ignore_certificates),
        }

        body_template = {}
        body_path = []

        if self.source_configuration.createRequest.body_options.object_path != "":
            body_template = self.source_configuration.createRequest.body
            body_path = (
                self.source_configuration.createRequest.body_options.object_path.split(
                    "."
                )
            )

        if self.source_configuration.createRequest.body_options.accepts_multiple:
            body = dict(body_template)
            allItems = []
            for item in data.iter_rows(named=True):
                allItems.append(item)

            if len(body_path) == 1:
                body[body_path[0]] = allItems
            elif len(body_path) == 2:
                body[body_path[0]][body_path[1]] = allItems
            elif len(body_path) == 3:
                body[body_path[0]][body_path[1]][body_path[2]] = allItems
            elif len(body_path) == 4:
                body[body_path[0]][body_path[1]][body_path[2]][body_path[3]] = allItems
            else:
                body = allItems

            request["json"] = body

            response = self.client.request(**request)
            if response.status_code > 299:
                raise Exception("Insertion failed: ", response.text)

        else:
            for item in data.iter_rows(named=True):
                body = dict(body_template)

                if len(body_path) == 1:
                    if body_path[0] == "0":
                        body = [item]
                    else:
                        body[body_path[0]] = item
                elif len(body_path) == 2:
                    if body_path[1] == "0":
                        body[body_path[0]] = [item]
                    else:
                        body[body_path[0]][body_path[1]] = item
                elif len(body_path) == 3:
                    if body_path[2] == "0":
                        body[body_path[0]][body_path[1]] = [item]
                    else:
                        body[body_path[0]][body_path[1]][body_path[2]] = item
                elif len(body_path) == 4:
                    if body_path[3] == "0":
                        body[body_path[0]][body_path[1]][body_path[2]] = [item]
                    else:
                        body[body_path[0]][body_path[1]][body_path[2]][
                            body_path[3]
                        ] = item
                else:
                    body = item

                if (
                    self.source_configuration.createRequest.body_type
                    == "application/json"
                ):
                    request["json"] = body
                else:
                    request["data"] = body

                response = self.client.request(**request)
                if response.status_code > 299:
                    print("Request failed: ", response.text)

    def insert(self, data: pl.DataFrame):
        return self._insert(data)
        raise NotImplementedError("Insert not implemented")

    def update(self, data: pl.DataFrame):
        raise NotImplementedError("Update not implemented")

    def delete(self, data: pl.DataFrame):
        if (
            self.connection_settings.soft_delete != True
            and self.source_configuration.deleteRequest.path == ""
        ):
            print("Delete not configured, skipping")
            return 0
        return 0



from hashlib import sha1
from typing import Any, Literal, Optional

import polars as pl
from bson import ObjectId

from .interface import (
    TransformerInterface,
    register_transformer,
    register_transformer_class,
)


@register_transformer_class("strings")
class StringTransformer(TransformerInterface):
    @staticmethod
    def staticfield(data: pl.DataFrame, field: str, value: str) -> pl.DataFrame:
        """Add a static field to the DataFrame

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(pl.Series(field, [value] * len(data)))
        return data

    @staticmethod
    def set_default(
        data: pl.DataFrame, inField: str, defaultValue: str
    ) -> pl.DataFrame:
        """Set a default value for a column if the current value is null

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            defaultValue (str): The default value

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(data[inField].fill_null(defaultValue))
        return data

    @staticmethod
    def strip(
        data: pl.DataFrame, inField: str, stripChars: str, outField: str = None
    ) -> pl.DataFrame:
        """Strip all given characters from a column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(
            data[inField]
            .str.strip(stripChars)
            .alias(outField if outField is not None else inField)
        )
        return data

    @staticmethod
    def lowercase(
        data: pl.DataFrame, inField: str = "", outField: str = "", inOutMap: dict = {}
    ) -> pl.DataFrame:
        """Transform all values in a column to lowercase and insert
            them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        if len(inOutMap) == 0:
            inOutMap = {
                inField: (
                    outField if outField is not None and outField != "" else inField
                )
            }

        for inf, outf in inOutMap.items():
            data = data.with_columns(data[inf].str.to_lowercase().alias(outf))

        return data

    @staticmethod
    def uppercase(
        data: pl.DataFrame, inField: str = "", outField: str = "", inOutMap: dict = {}
    ) -> pl.DataFrame:
        """Transform all values in a column to uppercase
            and insert them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        if len(inOutMap) == 0:
            inOutMap = {
                inField: (
                    outField if outField is not None and outField != "" else inField
                )
            }

        for inf, outf in inOutMap.items():
            data = data.with_columns(data[inf].str.to_uppercase().alias(outf))

        return data

    @staticmethod
    def match_contains(
        data: pl.DataFrame, inField: str, match: str, outField: str = ""
    ) -> pl.DataFrame:
        """Match a value in a column and insert a boolean value in another column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            matchValue (str): The value to match
            outField (str): The field to put the result in
        """

        return data.with_columns(
            data[inField]
            .str.contains(match)
            .alias(outField if outField is not None and outField != "" else inField)
        )

    @staticmethod
    def join(
        data: pl.DataFrame, inFields: list, outField: str, separator: str = ""
    ) -> pl.DataFrame:
        """Join multiple columns together

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inFields (list): The fields to join
            outField (str): The field to put the result in
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        nCol = pl.concat_str(data[inFields], separator=separator).alias(outField)

        return data.with_columns(nCol.alias(outField))

    @staticmethod
    def join_listfield(
        data: pl.DataFrame, inField: str, outField: str, separator: str = ","
    ) -> pl.DataFrame:
        try:
            if isinstance(type(data[inField].dtype), type(pl.List)):
                data = data.with_columns(
                    data[inField].list.join(separator).alias(outField)
                )
                return data

            data = data.with_columns(data[inField].arr.join(separator).alias(outField))
            return data

        except Exception:
            data = data.with_columns(
                data[inField].cast(pl.List).arr.join(separator).alias(outField)
            )

            return data

    @staticmethod
    def split(
        data: pl.DataFrame, inField: str, outFields: list, separator: str = ""
    ) -> pl.DataFrame:
        """Split a column into multiple fields

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to split
            outField (str): The fields to put the result in (in order)
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        amount = len(outFields)
        data = data.with_columns(
            data[inField]
            .str.splitn(separator, amount)
            .struct.unnest()
            .rename({f"field_{i}": fld for i, fld in enumerate(outFields)})
        )

        return data

    @staticmethod
    def split_into_listfield(
        data: pl.DataFrame, inField: str, outField: str = None, separator: str = ""
    ) -> pl.DataFrame:
        """Split a column into list field

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to split
            outField (str): The field to put the result into
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, [inField])

        old_column = data[inField]
        new_column = old_column.str.split(separator).alias(outField or inField)

        data = data.with_columns(new_column)

        return data

    @staticmethod
    def quote(data: pl.DataFrame, inField: str, outField: str = None, quote: str = "'"):
        """Quotes the given column values"""

        tmpData = StringTransformer.staticfield(data, "ccQuoteFld", quote)

        new = StringTransformer.join(
            tmpData, ["ccQuoteFld", inField, "ccQuoteFld"], outField, separator=""
        )

        return new

    @staticmethod
    def replace(
        data: pl.DataFrame,
        search: str,
        replace: str,
        inField: str,
        outField: str = None,
    ):
        """Replace a section in a string with another section"""

        return data.with_columns(
            data[inField]
            .str.replace(search, replace)
            .alias(outField if outField is not None else inField)
        )

    @staticmethod
    def replace_all(
        data: pl.DataFrame,
        search: str,
        replace: str,
        inField: str,
        outField: str = None,
    ):
        """Replace all matching sections in a string with another section"""

        return data.with_columns(
            data[inField]
            .str.replace_all(search, replace)
            .alias(outField if outField is not None else inField)
        )

    @staticmethod
    def substring(
        data: pl.DataFrame,
        inField: str,
        outField: str = None,
        start: int = 0,
        length: int = None,
    ):
        """Returns a substring of the given string column"""

        return data.with_columns(
            data[inField]
            .str.slice(start, length)
            .alias(outField if outField is not None else inField)
        )

    @staticmethod
    def add_prefix(data: pl.DataFrame, inField: str, prefix: str, outField: str = None):
        """Add a prefix to the given column"""

        return data.with_columns(
            pl.concat_str(
                pl.Series("ccTempField", [prefix] * len(data)), data[inField]
            ).alias(outField if outField is not None else inField)
        )

    @staticmethod
    def cast(data: pl.DataFrame, inField: str, outField: str = ""):
        """Cast a column to a different type"""

        if outField == "":
            outField = inField

        return data.with_columns(data[inField].cast(pl.Utf8).alias(outField))

    @staticmethod
    def to_object_id(data: pl.DataFrame, inField: str) -> pl.DataFrame:
        """Convert the given field to a bson.ObjectId

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to convert

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        data = data.with_columns(
            data[inField].map_elements(
                lambda oid: ObjectId(oid), return_dtype=pl.Object
            )
        )

        return data

    @staticmethod
    def format(
        data: pl.DataFrame,
        inField: str,
        outField: str = "",
        format_string: str = "{value}",
    ):
        """
        Formats the value in a string

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to format
            outField (str): The field to put the result in
            format_string (str): The format string using {value} as the placeholder for the value. Defaults to "{value}". Example: "_{value}_.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        if not outField:
            outField = inField

        data = data.with_columns(
            data[inField]
            .map_elements(
                lambda val: format_string.format(**{"value": val}), return_dtype=pl.Utf8
            )
            .alias(outField)
        )
        return data

    @staticmethod
    def hash(
        data: pl.DataFrame,
        outField: str,
        inField: Optional[str] = None,
        inFields: list[str] = [],
        hashWhen: Literal[
            "always", "any-value-is-populated", "all-values-are-populated"
        ] = "always",
    ):
        """
        Generates a hash from the concatinated field values

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to hash (if not inFields)
            inFields (list[str]): The fields to hash
            outField (str): The field to put the result n
            hashWhen ("always", "any-value-is-populated", "all-values-are-populated"): When to return a hashed value instead of none. Defaults to "always".

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        if inField:
            inFields.append(inField)

        def hash_function(fields: dict[str, Any]) -> str:
            for key, value in fields.items():
                if value is None:
                    fields[key] = ""

            concatenated_fields = "".join(fields.values())

            if hashWhen == "all-values-are-populated" and "" in fields.values():
                return None
            if hashWhen == "any-value-is-populated" and concatenated_fields == "":
                return None

            return sha1(concatenated_fields.encode("utf-8")).hexdigest()

        transformed_dataframe = data.with_columns(
            pl.struct(inFields)
            .map_elements(hash_function, return_dtype=str)
            .alias(outField)
        )

        return transformed_dataframe



from typing import Annotated, Literal

from pydantic import Field

from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class XmlConnectionArguments(SourceConnectionArguments):
    path: Annotated[str, Field(min_length=1)]
    encoding: Annotated[str, Field(default="utf-8")]


class XmlConfigArguments(SourceConfigArguments):
    connection: XmlConnectionArguments
    type: Annotated[Literal["Xml"], Field(default="Xml")] = "Xml"


class XmlConfig(SourceConfig):
    """The connection configuration class used for XML file sources"""

    path: str
    encoding: str = "utf-8"

    def __init__(self, arguments: XmlConfigArguments):
        super().__init__(arguments)
        self.path = arguments.connection.path
        self.encoding = arguments.connection.encoding



import concurrent.futures


class RequestThreader:
    threads: int = None
    executor: concurrent.futures.ThreadPoolExecutor = None

    def __init__(self, threads: int = 10):
        print("Starting threader...")
        self.threads = threads

    def __enter__(self):
        print("Threading the needle...")
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.threads)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        print("Waiting for threads to finish...")
        self.executor.shutdown(wait=True)

    def submit(self, func, *args, **kwargs):
        print("Submitting single thread...")
        return self.executor.submit(func, *args, **kwargs)

    def submitAndWait(self, func, kwarg_list: list):
        print("Submitting a list of threads...")
        for result in self.executor.map(lambda fn: func(**fn), kwarg_list):
            yield result



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class ExcelConnectionArguments(SourceConnectionArguments):
    path: Annotated[str, Field(min_length=1)]
    
class ExcelConfigArguments(SourceConfigArguments):

    type: Annotated[Literal["Excel"], Field(default="Excel")] = "Excel"
    connection: ExcelConnectionArguments


class ExcelConfig(SourceConfig):
    """The connection configuration class used for static sources"""

    path: str

    def __init__(self, arguments: ExcelConfigArguments):
        super().__init__(arguments)
        self.path = arguments.connection.path
